import time
import logging

import asyncio
from threading import Thread

import pika


class RabbitMQAsyncClient(object):
    """
    This client will handle rabbit MQ connections, failures and will allow for asynchronous message
    sending and receiving.

    It's supposed to be handy and very simple, so it won't let you create any exchange or channel.

    Based on pika package examples.
    """

    def __init__(self, parameters, on_message_callback, on_channel_callback, log_level=None, log_prefix="[RMQ] "):
        """
        :param pika.ConnectionParameters parameters: Connection parameters
        :param function on_message_callback(text): Function called when a message is received
        :param function on_channel_callback(RabbitMQAsyncClient instance): Function called when connection to RabbitMQ is established and channel is opened
        :param logging logger: Should accept .debug(), .info() and .error() methods (default: python logging)
        :param str log_prefix: A string displayed a the beginning of log message (default: "[RMQ] ")
        """
        self.connection = None
        self.channel = None
        self.closing = False
        self.consuming = False
        self.params = parameters

        fmt = ' '.join((log_prefix, '%(asctime)s - [%(levelname)s] %(message)s'))
        logging.basicConfig(level=log_level if log_level else logging.DEBUG, format=fmt)
        self.log = logging.getLogger(__name__)

        self.rcv_callback = on_message_callback
        self.channel_callback = on_channel_callback

    def connect(self):
        """ Open connection and handle disconnections. """
        self.log.info("Connecting to RabbitMQ server.")
        self.connection = pika.SelectConnection(
            parameters=self.params,
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_error,
            on_close_callback=self.on_connection_closed
        )
        self.run_ioloop_thread()

    def run_ioloop_thread(self):
        self.loop = Thread(target=self.run_ioloop)
        self.loop.start()

    def run_ioloop(self):
        self.connection.ioloop.start()

    def on_connection_open(self, connection):
        """ Open channel when connection is opened. """
        self.log.info("Connection opened.")
        self.open_channel()

    def on_connection_error(self, connection, error):
        """ Called when connection can't be opened. """
        self.log.error("Cannot connect: {}. Retrying in 5s.".format(error))
        self.connection.add_timeout(5, self.connect)

    def on_connection_closed(self, connection, reply_code, reply_text):
        """
        Notify disconnection. We cannot set a timer to reconnect because the ioloop will be stopped.

        :param pika.connection.Connection connection: The closed connection object
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given
        """
        self.channel = None
        self.log.info("Connection closed: ({0}) {1}.".format(reply_code, reply_text))

    def disconnect(self):
        """ Stop ongoing operations and disconnect from RabbitMQ server. """
        self.closing = True
        if self.consuming:
            self.stop_consuming()
        else:
            self.close_channel()
            self.connection.close()

    def open_channel(self):
        self.log.info("Opening a new channel.")
        self.connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        self.log.info("Channel opened.")
        self.channel = channel
        self.channel.add_on_close_callback(self.on_channel_closed)
        if self.channel_callback:
            self.channel_callback(self)

    def on_channel_closed(self, channel, reply_code, reply_text):
        """
        Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Will try to reopen later.

        :param pika.channel.Channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed

        """
        self.log.warning("Channel {0} closed: ({1}) {2}. Retrying in 5s.".format(channel.channel_number, reply_code, reply_text))
        self.connection.add_timeout(5, self.open_channel)
        self.channel = None
        self.consuming = False
        # self.connection.close()

    def close_channel(self):
        if self.channel:
            self.channel.close()
            self.log.info("Closing channel.")

    def start_consuming(self, queue):
        """
        Receive incoming messages.

        :param str queue: Queue from which to receive messages
        """
        self.log.info("Now consuming queue {}.".format(queue))
        self.channel.add_on_cancel_callback(self.on_consumer_cancelled)
        self.consumer_tag = self.channel.basic_consume(self.on_message, queue)
        self.consuming = True

    def on_message(self, channel, basic_deliver, properties, body):
        """
        Triggered when a message is received.
        This function is a first filter before calling user callback with less complexity.

        :param pika.channel.Channel channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param str|unicode body: The message body

        """
        self.log.debug("Received message '{0}'.".format(body))
        channel.basic_ack(basic_deliver.delivery_tag)
        self.rcv_callback(body)

    def on_consumer_cancelled(self, method_frame):
        """
        Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame
        """
        self.log.info("Consumer was cancelled remotely: {0}.".format(method_frame))
        if self.channel:
            self.channel.close()
        self.consuming = False

    def stop_consuming(self):
        """ Stop receiving incoming messages. """
        if self.channel:
            self.channel.basic_cancel(self.on_cancelok, self.consumer_tag)

    def on_cancelok(self, rabbitmq_cancel_frame):
        """ Triggered when RabbitMQ acknowledges consuming cancelation. """
        self.log.info("Sucessfully canceled consuming.")
        self.consuming = False
        if self.closing:
            self.close_channel()
            self.connection.close()

    def send(self, exchange, key, message):
        """
        Send a message.

        :param str exchange: The exchange
        :param str key: The routing key
        :param str message: The message body (text)

        :return bool: Whether the message was sent or not
        """
        self.log.debug("sending " + str(message))
        if not self.channel:
            self.log.warning("Cannot send message: channel is not opened!")
            return False
        try:
            self.channel.basic_publish(exchange=exchange, routing_key=key, body=message)
            self.log.debug("Sent message '{0}' through exhange {1} with key {2}.".format(message, exchange, key))
            return True
        except pika.exceptions.ConnectionClosed as e:
            self.log.error("Cannot send message: connection is closed. Reconnecting and sending again in 5s.")
            self.connect()
            self.connection.add_timeout(5, self.send, [exchange, key, message])
            pass
        return False

    def on_delivery_confirmation(self, method_frame):
        """
        UNUSED
        Invoked when message destination answers to incoming message.

        :param pika.frame.Method method_frame: Basic.Ack or Basic.Nack frame
        """
        confirmation_type = method_frame.method.NAME.split('.')[1].lower()
        self.log.debug(str(method_frame.method.NAME))
        if confirmation_type == 'ack':
            self.log.debug("Message received by destination.")
        elif confirmation_type == 'nack':
            self.log.warning("Message was not received by destination.")



import sys
import time
import uuid
import json
import pika
import datetime
import itertools

import logging

logging.getLogger('pika').setLevel(logging.INFO)

logger = logging.getLogger('hotot')


class BaseClient(object):
    def __init__(self, host, exchange='', header={}, timeout=60, connection_attempts=100, logging_level=None):
        self.timeout = timeout
        self.exchange = exchange
        self.header = header
        self.response = None

        if logging_level:
            logging.getLogger('pika').setLevel(logging_level)

        self.connection = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=host,
                connection_attempts=connection_attempts,
                retry_delay=5
            )
        )

        self.channel = self.connection.channel()

    def _publish(self, routing_key, props, mandatory=False, persistent=False, **kwargs):
        return self.channel.basic_publish(
            exchange=self.exchange,
            routing_key=routing_key,
            properties=pika.BasicProperties(**{
                **props,
                **{
                    'content_type': 'application/json',
                    'delivery_mode': 2 if persistent else 1
                }
            }),
            mandatory=mandatory,
            body=json.dumps({**self.header, **kwargs})
        )

    def exit(self):
        self.channel.stop_consuming()
        self.connection.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.exit()


class RetryException(Exception):
    pass


class StopRetryException(Exception):
    pass


def ack(func):
    def func_wrapper(ch, method, properties, body):
        try:
            func(ch, method, properties, json.loads(body))
            ch.basic_ack(method.delivery_tag)
        except Exception as e:
            logger.error('{}'.format(e))
            ch.basic_reject(method.delivery_tag, requeue=False)
    return func_wrapper


def retry(exchange, max_retries):
    def _retry(func, error):
        def func_wrapper(ch, method, properties, body):
            try:
                json_body = json.loads(body)
                func(ch, method, properties, json_body)
                ch.basic_ack(method.delivery_tag)
            except RetryException:
                ch.basic_reject(method.delivery_tag, requeue=False)
                count = 0
                if properties.headers:
                    count = properties.headers['x-death'][0]['count']
                logger.warning('({}/{}) Rerouting message {} {}'.format(count, max_retries, method.routing_key, body))
                if max_retries and count >= max_retries:
                    error(ch, method, properties, json_body)
                    return
                else:
                    ch.basic_publish(exchange=exchange,
                                     routing_key=method.routing_key,
                                     properties=properties,
                                     body=body)
            except StopRetryException:
                ch.basic_reject(method.delivery_tag, requeue=False)
                error(ch, method, properties, json_body)
                return
        return func_wrapper
    return _retry


def default_callback_max_retries(ch, method, properties, body):
    print('Stop retrying...')


class ConsumerClient(BaseClient):
    def __init__(self,
                 host,
                 exchange='',
                 queue='',
                 exchange_type='direct',
                 routing_key=None,
                 exclusive=False,
                 durable=False,
                 retry_ttl=None,
                 max_retries=None,
                 **kwargs):

        super().__init__(host, exchange, **kwargs)

        self.queue = queue
        self.channel.exchange_declare(self.exchange, exchange_type=exchange_type, durable=durable)
        result = self.channel.queue_declare(queue=self.queue, exclusive=exclusive, durable=durable)
        self.channel.queue_bind(exchange=self.exchange, queue=result.method.queue, routing_key=routing_key)

        self.retry_ttl = retry_ttl
        if self.retry_ttl:
            self.max_retries = max_retries
            self.dlx = 'dlx-' + self.exchange

            self.channel.exchange_declare(self.dlx, exchange_type='fanout', auto_delete=True)
            result = self.channel.queue_declare(
                queue='dlq-' + self.queue,
                auto_delete=True,
                arguments={
                    'x-message-ttl': self.retry_ttl,
                    'x-dead-letter-exchange': self.exchange
                }
            )
            self.channel.queue_bind(exchange=self.dlx, queue=result.method.queue)

    def consume(self, callback, callback_max_retries=None):
        """ Consume messages from queue using a callback """
        if self.retry_ttl:
            if not callback_max_retries:
                callback_max_retries = default_callback_max_retries
            rt = retry(self.dlx, self.max_retries)
            callback = rt(callback, callback_max_retries)
        else:
            callback = ack(callback)

        self.channel.basic_consume(callback, queue=self.queue)
        self.channel.start_consuming()


class ProducerClient(BaseClient):

    def __init__(self, host, exchange='', ack_receipt=False, max_retries=None, **kwargs):
        super().__init__(host, exchange, **kwargs)
        self.ack_receipt = ack_receipt
        self.max_retries = max_retries
        if self.ack_receipt:
            self.channel.confirm_delivery()

    def publish(self, routing_key, **kwargs):
        """ Publish a message on the output exchange """
        if self.ack_receipt:
            retries = range(self.max_retries) if self.max_retries else itertools.repeat(None)
            for i in retries:
                if self._publish(routing_key, {}, mandatory=True, **kwargs):
                    return True
            return False
        else:
            self._publish(routing_key, {}, **kwargs)


class RpcClient(BaseClient):

    def _on_response(self, ch, method, props, body):
        """ Filter incoming responses on correlation ID """
        if not(hasattr(self, 'corr_id')) or self.corr_id == props.correlation_id:
            self.response = json.loads(body)

    def _wait(self, frequency=0, timeout=None):
        """ Check for a response at a specific frequency during a defined duration """
        t_end = time.time() + (timeout if timeout else self.timeout)
        while self.response is None:
            time.sleep(frequency)
            if time.time() > t_end:
                break
            self.connection.process_data_events()

    def _open_callback(self, queue=''):
        """ Declare and start the callback queue if needed
            If no queue name is provided, a random name is chosen and the queue is exclusive
        """
        if not hasattr(self, 'callback_queue'):
            result = self.channel.queue_declare(queue=queue, exclusive=not(bool(queue)))
            self.callback_queue = result.method.queue
            self.channel.basic_consume(
                self._on_response,
                no_ack=True,
                queue=self.callback_queue
            )

    def call(self, routing_key, **kwargs):
        """ Publish a message and await for the response """
        self.response = None
        self._open_callback()

        self.corr_id = str(uuid.uuid4())
        self._publish(
            routing_key, {
                'reply_to': self.callback_queue,
                'correlation_id': self.corr_id
            }, **kwargs)

        self._wait()
        return self.response

    def send(self, routing_key, **kwargs):
        """ Publish a message on the output exchange without a waiting block """
        return self._publish(routing_key, {}, **kwargs)

    def wait(self, frequency=0, timeout=None):
        """ Wait for a response on the callback queue """
        self.response = None
        self._wait(frequency, timeout)
        return self.response



import os
import logging
import requests
from functools import partial

methods = {'GET': requests.get, 'POST': requests.post, 'PUT': requests.put}


def _api(method, host, port, logger, req, mask=None, params={}, **kwargs):
    try:
        if not host.startswith('http'):
            host = 'http://' + host
        api_url = '{}{}/{}'.format(host, ':{}'.format(port) if port else '', req)
        response = methods[method](api_url, params=params, **kwargs)
        code = response.status_code
        req_msg = '[{}] API request to {}:'.format(code, response.url)
        if code in [200, 201]:
            if code == 200:
                logger.info('{} Successful response!'.format(req_msg))
            elif code == 201:
                logger.info('{} Ressource created!'.format(req_msg))

            response = response.json()
            if mask:
                if isinstance(mask, str):
                    return [elt[mask] for elt in response]
                return [{k: v for k, v in elt.items() if k in mask} for elt in response]
            return response
        else:
            if code >= 500:
                msg = 'Server Error'
            elif code == 404:
                msg = 'URL not found'
            elif code == 401:
                msg = 'Authentication Failed'
            elif code == 400:
                msg = 'Bad Request'
            elif code >= 300:
                msg = 'Unexpected Redirect'
            logger.error('{} {}'.format(req_msg, msg))
            return None
    except requests.exceptions.ConnectionError:
        logger.exception('Connection error')
        return None


def api(method='GET', host='localhost', port='', logger=None, api_token=None):
    logging.basicConfig(level=logging.ERROR)
    logger = logger or logging.getLogger(__name__)
    return partial(_api, method,
                   os.environ.get(str(host), host),
                   os.environ.get(str(port), port),
                   logger,
                   headers={
                       **({'Authorization': 'Bearer {}'.format(api_token)} if api_token else {})
                   }
                   )



import os
import setuptools

with open('INFO.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name=os.environ.get('CI_PROJECT_NAME', 'hotot'),
    version=os.environ.get('CI_COMMIT_TAG', os.environ.get('VERSION', '') + 'dev'),
    author='Lancey Energy Storage',
    description='RabbitMQ/Pika Python client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=os.environ.get('CI_PROJECT_URL', ''),
    packages=setuptools.find_packages(),
    classifiers=(
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'License :: Public Domain'
    ),
    install_requires=['pika', 'requests'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)
